import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ezdnac",
    version="0.0.48",
    author="Johan Lahti",
    author_email="ccie60702@gmail.com",
    description="A small module to make it easy to start using the rest-api of Cisco DNA-C",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/johan-lahti/ezdnac",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=['requests','tabulate', 'click'],
    python_requires='>=3.6',
)



from ezdnac.excepts import *
from ezdnac.utils import *
import requests
import json
import re
import os

class Dnac:
    def __init__(self, ip, uid, pw=None, **kwargs):
        self.authToken = None
        self.timeout = 5
        self.ip = ip
        self.uid = uid
        self.pw = pw
        self.baseurl = '/api/v1/'
        self.authBaseUrl = '/api/system/v1/'
        self.taskId = None
        # Setting possible keyword adguments
        # Port
        if 'port' in kwargs:
            self.port = kwargs['port']
        else:
            self.port = "443"
        # Verify SSL
        if 'verifySSL' in kwargs:
            verifySSL = kwargs['verifySSL']
            self.verifySSL = kwargs['verifySSL']
        else:
            verifySSL = False
            self.verifySSL = False
        # Authtoken
        if 'authToken' in kwargs:
            self.authToken = kwargs['authToken']
            authToken = kwargs['authToken']
        # Timeout
        if self.timeout is None:
            try:
                timeout = kwargs['timeout']
                self.timeout = timeout
            except KeyError:
                timeout = 5
                self.timeout = timeout
            else:
                self.timeout = 5
        # Auth
        self.auth()

    def auth(self):
        if self.authToken is None:
            print("Authenticating..")
            # authenticate with username/password and populate token
            AuthURL = "https://" + self.ip + ":" + \
                self.port + self.authBaseUrl + "auth/token"
            headers = {}
            try:
                response = requests.post(AuthURL, headers=headers, verify=self.verifySSL, auth=(
                    self.uid, self.pw), timeout=self.timeout)
            except:
                print(
                    "Error: Timeout connection to DNA-C. Most likely a network reachability issue")
                exit()
            data = json.loads(response.text)
            if 'error' in data:
                print (data['error'])
                exit()
            else:
                print ("Login Success")
            authToken = (data['Token'])
            self.authToken = authToken
            return self.authToken
        elif self.authToken is not None:
            print ("Reusing existing key..")
            return self.authToken

    def reauth(self):
        print("Authenticating..")
        # authenticate with username/password and populate token
        AuthURL = "https://" + self.ip + ":" + \
            self.port + self.authBaseUrl + "auth/token"
        headers = {}
        try:
            response = requests.post(AuthURL, headers=headers, verify=self.verifySSL, auth=(
                self.uid, self.pw), timeout=self.timeout)
        except:
            print(
                "Error: Timeout connection to DNA-C. Most likely a network reachability issue")
            exit()
        data = json.loads(response.text)
        if 'error' in data:
            print (data['error'])
            exit()
        else:
            print ("Login Success")
        authToken = (data['Token'])
        self.authToken = authToken
        return self.authToken

    def taskStatus(self, **kwargs):
        try:
            self.taskId = kwargs['id']
        except:
            pass

        if self.taskId == None:
            raise ezDNACError('No previous task to check')

        endpoint = f'task/{self.taskId}'

        data = restcall('GET', self, endpoint)
        return data['response']

    # Get the selected device ID from serial:

    def getAllDevices(self):
        endpoint = "network-device/"
        data = restcall('GET', self, endpoint)

        return data

    def id_from_serial(self, serialNumber):
        switches = self.getAllDevices()
        for switch in switches['response']:
            if switch['serialNumber'] == serialNumber:
                switchId = switch['id']
        try:
            return switchId
        except KeyError:
            return None

    def getTemplates(self):
        endpoint = "template-programmer/project"
        data = restcall('GET', self, endpoint)

        return data

    def getTemplateId(self, templateName):
        data = self.getTemplates()
        for projects in data:
            for templates in projects['templates']:
                try:
                    if templates['name'] == templateName:
                        return (templates['id'])
                except:
                    return None

    def getTemplateInfo(self, templateId):
        endpoint = f'template-programmer/template/{templateId}'
        data = restcall('GET', self, endpoint)

        return data

    def getSites(self, **kwargs):
        try:
            searchsite = kwargs['site']
        except:
            searchsite = None
        if searchsite == None:
            url = "https://" + self.ip + ":" + self.port + "/dna/intent/api/v1/site"
        else:
            url = "https://" + self.ip + ":" + self.port + \
                "/dna/intent/api/v1/site?name=" + searchsite + ""

        payload = {}
        headers = {
            'x-auth-token': self.authToken,
            'Content-Type': 'application/json',
            '__runsync': 'true',
            '__timeout': '10',
            '__persistbapioutput': 'true',
        }

        response = requests.request(
            "GET", url, headers=headers, data=payload, verify=self.verifySSL, timeout=self.timeout)
        data = json.loads(response.text)

        if searchsite != None:
            for site in data['response']:
                return site['id']

        return data['response']

    def getPnpDevices(self, **kwargs):
        if 'sn in kwargs':
            serialNumber = kwargs['sn']
        else:
            serialNumber = None

        endpoint = "onboarding/pnp-device"
        payload = {}
        headers = {
            'x-auth-token': self.authToken
        }
 #       response = requests.request(
 #           "GET", url, headers=headers, data=payload, verify=self.verifySSL, timeout=self.timeout)
 #       data = json.loads(response.text)


        response = restcall('GET', self, endpoint)
        data = response

        if serialNumber is None:
            return data

        else:
            for device in data:
                if serialNumber in device['deviceInfo']['serialNumber']:
                    return device

            # If sn was set, but not found:
            raise ezDNACError('Device with serial number' +
                               serialNumber + ' not found in pnp.')

    def getInventoryDevies(self, **kwargs):

        # Setting attributes from kwargs
        if 'sn' in kwargs:
            serialNumber = kwargs['sn']
        else:
            serialNumber = None

        if 'id' in kwargs:
            deviceId = kwargs['id']
        else:
            deviceId = None

        if 'hostname' in kwargs:
            hostname = kwargs['hostname']
        else:
            hostname = None

        # Getting data from dnac
        # If device-id is used
        if deviceId is not None:
            endpoint = f"network-device/{deviceId}"

            response = restcall('GET', self, endpoint)
            data = response['response']
            return data

        # If serialNumber is used
        elif serialNumber is not None:
            endpoint = "network-device/"

            response = restcall('GET', self, endpoint)
            data = response['response']

            for device in data:
                if 'serialNumber' in device:
                    if serialNumber in device['serialNumber']:
                        return device


        # If hostname is used
        elif hostname is not None:
            endpoint = "network-device/"
            response = restcall('GET', self, endpoint)
            data = response['response']
            for device in data:
                if re.match(rf'{hostname}.*', device['hostname']):
                    return device

    def pullTemplates(self, **kwargs):
        path = ""
        projectName = None
        try:
            projectName = kwargs['project']
        except:
            pass
        try:
            path = kwargs['path']
        except:
            pass

        url = "https://" + self.ip + ":" + self.port + \
            baseurl + "template-programmer/project"
        headers = {
            'x-auth-token': self.authToken,
            'Content-Type': 'application/json',
        }
        response = requests.request(
            "GET", url, headers=headers, verify=self.verifySSL, timeout=self.timeout)
        data = json.loads(response.text)
        templateslist = []
        # Get the id of interesting templates:
        for projects in data:
            if projectName != None:
                if (projects['name']) == projectName:
                    for template in projects['templates']:
                        templates = {}
                        templates['id'] = template['id']
                        templateslist.append(templates)
            else:
                for template in projects['templates']:
                    templates = {}
                    templates['id'] = template['id']
                    templateslist.append(templates)

        for template in templateslist:
            url = "https://" + self.ip + ":" + self.port + baseurl + \
                "template-programmer/template/" + template['id']
            headers = {
                'x-auth-token': self.authToken,
                'Content-Type': 'application/json',
            }
            response = requests.request(
                "GET", url, headers=headers, verify=self.verifySSL, timeout=self.timeout)
            templateData = json.loads(response.text)

            # Create template path folder if not exists already
            if not os.path.exists(path):
                os.mkdir(path)

            # Creates one subfolder per template, containting separate files for parameters and content
            templatePath = path + templateData['name']
            if not os.path.exists(templatePath):
                os.mkdir(templatePath)

            # Create a file for the actual content:
            contentsFilename = templatePath + "/" + \
                templateData['name'] + "_contents.txt"
            with open(contentsFilename, 'w') as out:
                templateParams = templateData['templateContent']
                out.write(str(templateParams))

            # Create a file for all parameters:
            paramsFilename = templatePath + "/" + \
                templateData['name'] + "_params.json"
            with open(paramsFilename, 'w') as out:
                templateParams = templateData
                if 'templateContent' in templateParams:
                    del templateParams['templateContent']
                out.write(str(json.dumps(templateParams, indent=4)))

        if path == "":
            path = "local folder"
        if projectName != None:
            return "All templates in project: " + projectName + " are synced to: " + path
        else:
            return "All templates in all projects are synced to: " + path

    def pushTemplates(self, **kwargs):
        path = ""
        try:
            path = kwargs['path']
        except:
            pass

        # Createa a reference dict for each templatefolder:
        templatesList = []
        listdir = os.listdir(path)
        folders = []
        for obj in listdir:
            if not re.match(r"^\.", obj):
                folders.append(obj)

        for folder in folders:
            templateDict = {"name": folder}
            templateName = folder
            templatePath = path + folder

            for file in os.listdir(path + folder):
                if re.match(r'.*_params.json', file):
                    templateDict['paramsFile'] = templatePath + "/" + file

                if re.match(r'.*_contents.txt', file):
                    templateDict['contentsFile'] = templatePath + "/" + file

            templatesList.append(templateDict)

        # Firest check if template already exists:
        for template in templatesList:

            # Opening both files
            # Content from the _content.json file:
            templateContents = str(open(template['contentsFile'], 'r').read())

            # All parameters from the params file:
            with open(template['paramsFile']) as templateData:
                templateFile = json.load(templateData)
                templateName = templateFile['name']
                projectName = templateFile['projectName']

                # Check whats already existing, based on project/template tree name
                url = "https://" + self.ip + ":" + self.port + \
                    baseurl + "template-programmer/project/"
                payload = {}
                headers = {
                    'x-auth-token': self.authToken,
                    'Content-Type': 'application/json',
                }
                response = requests.request(
                    "GET", url, headers=headers, data=payload, verify=self.verifySSL, timeout=self.timeout)
                data = json.loads(response.text)
                templateExists = False
                projectExists = False
                for project in data:
                    if project['name'] == projectName:
                        projectExists = True
                        projectId = project['id']
                        print ("Project exists: " +
                               projectName + " - " + projectId)

                        for templates in project['templates']:
                            if templates['name'] == templateName:
                                templateId = templates['id']
                                templateExists = True
                                print ("Template exists: " +
                                       templateName + " - " + templateId)

                if projectExists == False:
                    print ("Creating missing project: " + projectName)
                    url = "https://" + self.ip + ":" + self.port + \
                        "/dna/intent/api/v1/template-programmer/project"

                    payload = {
                        "name": projectName,
                    }
                    headers = {
                        'x-auth-token': self.authToken,
                        'Content-Type': 'application/json',
                    }
                    response = requests.request(
                        "POST", url, headers=headers, json=payload, verify=self.verifySSL, timeout=self.timeout)
                    data = json.loads(response.text)

                    # Since the project didn't exist, we need to fetch it's new id.
                    taskId = data['response']['taskId']
                    datafromtask = self.taskStatus(id=taskId)
                    projectId = datafromtask['data']

                if templateExists == False:
                    print ("Creating missing template: " + templateName)

                    url = "https://" + self.ip + ":" + self.port + \
                        "/dna/intent/api/v1/template-programmer/project/" + projectId + "/template"
                    payload = templateFile
                    # Remove keys, making the payload suitable for new-creation of template. Hence removing id etc.
                    if 'id' in payload:
                        del payload['id']
                    for param in payload['templateParams']:
                        if 'id' in param:
                            del param['id']

                    for param in payload['templateParams']:
                        if 'selection' in param:
                            del param['selection']

                    payload['templateContent'] = str(templateContents)

                    headers = {
                        'x-auth-token': self.authToken,
                        'Content-Type': 'application/json',
                    }
                    response = requests.request(
                        "POST", url, headers=headers, json=payload, verify=self.verifySSL, timeout=self.timeout)
                    data = json.loads(response.text)
                    try:
                        self.taskId = data['response']['taskId']
                    except:
                        self.taskId = None

                if templateExists == True:
                    # Since the template exists, PUT the file from directory to make sure the active template is same version.
                    print ("Template " + templateName +
                           " already exists, versioning and updating.")

                    headers = {
                        'x-auth-token': self.authToken,
                        'Content-Type': 'application/json',
                    }

                    versionUrl = "https://" + self.ip + ":" + self.port + \
                        baseurl + "template-programmer/template/version"
                    versionPayload = {
                        'comments': 'Updated with EZDNAC',
                        'templateId': templateId
                    }
                    requests.post(versionUrl, headers=headers, json=versionPayload,
                                  verify=self.verifySSL, timeout=self.timeout)

                    url = "https://" + self.ip + ":" + self.port + \
                        baseurl + "template-programmer/template/"
                    payload = templateFile
                    payload['id'] = templateId

                    for params in payload['templateParams']:
                        del params['id']

                    for param in payload['templateParams']:
                        if 'selection' in param:
                            del param['selection']

                    payload['templateContent'] = templateContents

                    response = requests.request(
                        "PUT", url, headers=headers, json=payload, verify=self.verifySSL, timeout=self.timeout)
                    data = json.loads(response.text)



from ezdnac.excepts import *
from ezdnac.utils import *
import ezdnac.dnac
import requests
import json
import re
import os



class Template():
    def __init__(self, dnac, **kwargs):
        """
        Creates a new Template obj.

        Input parameters
        :dna(ezdnac DnaC object)
        :kwargs['all'] = Boolean # If True, all templates are choosen
        :kwargs['id'] = STRING # id of the template to work with
        :kwargs['name'] = STRING # name of the template
        """
        self.id = None
        self.name = None
        self.all = False
        self.data = None
        self.params = None
        self.authToken = dnac.authToken
        self.dnacIP = dnac.ip
        self.port = dnac.port
        self.uid = dnac.uid
        self.pw = dnac.pw
        self.dnac = dnac

        if 'all' in kwargs:
            if kwargs['all'] is True:
                self.all = True
                self.data = self.getTemplates()

        elif 'id' in kwargs:
            self.id = kwargs['id']
            allTemplates = self.getTemplates()
            for project in allTemplates:
                for template in project['templates']:
                    if template['id'] == kwargs['id']:
                        self.data = self.getTemplateInfo(self.id)
                        self.name = template['name']

        elif 'name' in kwargs:
            self.name = kwargs['name']
            allTemplates = self.getTemplates()
            for project in allTemplates:
                for template in project['templates']:
                    if template['name'] == kwargs['name']:
                        self.id = template['id']
                        self.data = self.getTemplateInfo(self.id)

    def setParams(self, params):
        if type(params) is dict:
            self.params = params
        else:
            raise ezDNACError('Template parameters has to be of type dict.')
        return "Params set"

    def getTemplates(self):
        endpoint = "template-programmer/project"
        data = restcall('GET', self.dnac, endpoint)
        return data

    def getTemplateId(self, templateName):
        data = self.getTemplates()
        for projects in data:
            for templates in projects['templates']:
                try:
                    if templates['name'] == templateName:
                        return (templates['id'])
                except:
                    return None

    def getTemplateInfo(self, templateId):
        endpoint = "template-programmer/template/" + templateId
        data = restcall('GET', self.dnac, endpoint)
        return data

    def newVersion(self):
        endpoint = "template-programmer/template/version"
        payload = {
            'comments': 'Updated with EZDNAC',
            'templateId': self.id
        }
        data = restcall('POST', self.dnac, endpoint, jsondata=payload)

    def pullTemplates(self, **kwargs):
        path = ""
        projectName = None
        try:
            projectName = kwargs['project']
        except:
            pass
        try:
            path = kwargs['path']
        except:
            pass

        endpoint = "template-programmer/project"
        data = restcall('GET', self.dnac, endpoint)

        templateslist = []
        # Get the id of interesting templates:
        for projects in data:
            if projectName is not None:
                if (projects['name']) == projectName:
                    for template in projects['templates']:
                        templates = {}
                        templates['id'] = template['id']
                        templateslist.append(templates)
            else:
                for template in projects['templates']:
                    templates = {}
                    templates['id'] = template['id']
                    templateslist.append(templates)

        for template in templateslist:
            url = "https://" + self.ip + ":" + self.port + baseurl + \
                "template-programmer/template/" + template['id']

            templateEndpoint = "template-programmer/template/" + template['id']
            templateData = restcall('GET', self.dnac, templateEndpoint)

            # Create template path folder if not exists already
            if not os.path.exists(path):
                os.mkdir(path)

            # Creates one subfolder per template, containting separate files for parameters and content
            templatePath = path + templateData['name']
            if not os.path.exists(templatePath):
                os.mkdir(templatePath)

            # Create a file for the actual content:
            contentsFilename = templatePath + "/" + \
                templateData['name'] + "_contents.txt"
            with open(contentsFilename, 'w') as out:
                templateParams = templateData['templateContent']
                out.write(str(templateParams))

            # Create a file for all parameters:
            paramsFilename = templatePath + "/" + \
                templateData['name'] + "_params.json"
            with open(paramsFilename, 'w') as out:
                templateParams = templateData
                if 'templateContent' in templateParams:
                    del templateParams['templateContent']
                out.write(str(json.dumps(templateParams, indent=4)))

        if path == "":
            path = "local folder"
        if projectName is not None:
            return "All templates in project: " + projectName + " are synced to: " + path
        else:
            return "All templates in all projects are synced to: " + path

    def pushTemplates(self, **kwargs):
        path = ""
        try:
            path = kwargs['path']
        except:
            pass

        # Createa a reference dict for each templatefolder:
        templatesList = []
        listdir = os.listdir(path)
        folders = []
        for obj in listdir:
            if not re.match(r"^\.", obj):
                folders.append(obj)

        for folder in folders:
            templateDict = {"name": folder}
            templateName = folder
            templatePath = path + folder

            for file in os.listdir(path + folder):
                if re.match(r'.*_params.json', file):
                    templateDict['paramsFile'] = templatePath + "/" + file

                if re.match(r'.*_contents.txt', file):
                    templateDict['contentsFile'] = templatePath + "/" + file

            templatesList.append(templateDict)
        print(templateDict)

        # Firest check if template already exists:
        for template in templatesList:

            # Opening both files
            # Content from the _content.json file:
            templateContents = str(open(template['contentsFile'], 'r').read())

            # All parameters from the params file:
            with open(template['paramsFile']) as templateData:
                templateFile = json.load(templateData)
                templateName = templateFile['name']
                projectName = templateFile['projectName']

                # Check whats already existing, based on project/template tree name
                endpoint = "template-programmer/project/"

                data = restcall('GET', self.dnac, endpoint)

                templateExists = False
                projectExists = False
                for project in data:
                    if project['name'] == projectName:
                        projectExists = True
                        projectId = project['id']
                        print ("Project exists: " +
                               projectName + " - " + projectId)

                        for templates in project['templates']:
                            if templates['name'] == templateName:
                                templateId = templates['id']
                                templateExists = True
                                print ("Template exists: " +
                                       templateName + " - " + templateId)

                if projectExists == False:
                    print ("Creating missing project: " + projectName)
                    endpoint = "template-programmer/project"

                    payload = {
                        "name": projectName,
                    }

                    data = restcall('POST', self.dnac, endpoint, json=payload)

                    # Since the project didn't exist, we need to fetch it's new id.
                    taskId = data['response']['taskId']
                    datafromtask = self.taskStatus(id=taskId)
                    projectId = datafromtask['data']

                if templateExists == False:
                    print ("Creating missing template: " + templateName)

                    endpoint = "template-programmer/project/" + projectId + "/template"
                    payload = templateFile
                    # Remove keys, making the payload suitable for new-creation of template. Hence removing id etc.
                    if 'id' in payload:
                        del payload['id']
                    for param in payload['templateParams']:
                        if 'id' in param:
                            del param['id']

                    for param in payload['templateParams']:
                        if 'selection' in param:
                            del param['selection']

                    payload['templateContent'] = str(templateContents)

                    data = restcall('POST', self.dnac, endpoint, json=payload)
                    
                    try:
                        self.taskId = data['response']['taskId']
                    except:
                        self.taskId = None

                if templateExists == True:
                    # Since the template exists, PUT the file from directory to make sure the active template is same version.
                    print ("Template " + templateName +
                           " already exists, versioning and updating.")

                    headers = {
                        'x-auth-token': self.authToken,
                        'Content-Type': 'application/json',
                    }

                    versionEndpoint = "template-programmer/template/version"
                    versionPayload = {
                        'comments': 'Updated with EZDNAC',
                        'templateId': templateId
                    }

                    restcall('POST', self.dnac, versionEndpoint, json=versionPayload)

                    endpoint = "template-programmer/template/"
                    payload = templateFile
                    payload['id'] = templateId

                    for params in payload['templateParams']:
                        del params['id']

                    for param in payload['templateParams']:
                        if 'selection' in param:
                            del param['selection']

                    payload['templateContent'] = templateContents

                    data = restcall('PUT', self.dnac, endpoint, json=payload)




from ezdnac.excepts import *
from ezdnac.utils import *
import requests
import json
import re
import os


# When initialized, populate device parameters:
# Retreive switchId based on serialnumber
class Device():
    def __init__(self, dnac, **kwargs):
        self.collectionStatus = None
        self.dnac = dnac
        self.id = None
        self.hostname = None
        self.serialNumber = None
        self.state = None
        self.deploymentId = None
        self.executionId = None
        self.taskId = None

        if 'id' in kwargs:
            self.id = kwargs['id']
            self.initMethod = 'id'
        elif 'sn' in kwargs:
            self.serialNumber = kwargs['sn']
            self.initMethod = 'sn'
        elif 'hostname' in kwargs:
            self.hostname = kwargs['hostname']
            self.initMethod = 'hostname'

        if self.id is None and self.serialNumber is None and self.hostname is None:
            raise ezDNACError(
                'No device argment found. Enter hostname, id or sn')

        # if init method is id, the device must be in inventory. Populate all attributes:
        if self.initMethod == 'id':
            try:
                self.state = "Provisioned"
                INVdevices = self.dnac.getInventoryDevies(id=self.id)
            except:
                raise ezDNACError('device not found by id')

        elif self.initMethod == 'hostname':
            try:
                INVdevices = self.dnac.getInventoryDevies(
                    hostname=self.hostname)
            except:
                raise ezDNACError('device not found by hostname')

        # if method is sn, the device can be either in inventory or pnp, have to try both
        elif self.initMethod == 'sn':

            # Try inventory:
            INVdevices = self.dnac.getInventoryDevies(sn=self.serialNumber)
            if INVdevices is not None:
                self.state = "Provisioned"
                self.id = INVdevices['id']
                self.serialNumber = INVdevices['serialNumber']
                self.ip = INVdevices['managementIpAddress']
                self.hostname = INVdevices['hostname']
                self.platform = INVdevices['platformId']
                self.softwareVersion = INVdevices['softwareVersion']
                self.softwareType = INVdevices['softwareType']
                self.collectionStatus = INVdevices['collectionStatus']


        if self.state != "Provisioned" and self.initMethod == 'sn':
            try:
                # Otherwise try populate attributes via pnp:
                PNPdevices = self.dnac.getPnpDevices(sn=self.serialNumber)

                # If stack, check first switch
                if 'stackInfo' in PNPdevices['deviceInfo']:
                    if 'stackMemberList' in PNPdevices['deviceInfo']['stackInfo']:
                        if len(PNPdevices['deviceInfo']['stackInfo']['stackMemberList']) > 1:
                            
                            self.hostname = PNPdevices['deviceInfo']['name']
                            self.id = PNPdevices['id']
                            self.softwareType = PNPdevices['deviceInfo']['agentType']
                            self.state = PNPdevices['deviceInfo']['state']
                            self.platform = PNPdevices['deviceInfo']['pid']

                            # Some attributes is picked from first switch in stack.                            
                            stack1 = PNPdevices['deviceInfo']['stackInfo']['stackMemberList'][0]
                            self.softwareVersion = stack1['softwareVersion']       
                else:
                    self.id = PNPdevices['id']
                    self.state = PNPdevices['state']
                    self.hostname = PNPdevices['name']
                    self.platform = PNPdevices['pid']
                    self.softwareVersion = PNPdevices['softwareVersion']
                    self.softwareType = PNPdevices['agentType']
            except:
                raise ezDNACError('device with serialNumber ' +
                                  str(self.serialNumber) + ' not found')
            try:
                httpHeaders = PNPdevices['deviceInfo']['httpHeaders']
                for header in httpHeaders:
                    if header['key'] == 'clientAddress':
                        self.ip = header['value']
            except:
                pass

    def updateAttributes(self):
        endpoint = f"network-device/{self.id}"
        response = restcall('GET', self.dnac, endpoint)['response']

        self.serialNumber = response['serialNumber']
        self.ip = response['managementIpAddress']
        self.hostname = response['hostname']
        self.platform = response['platformId']
        self.softwareVersion = response['softwareVersion']
        self.softwareType = response['softwareType']
        self.collectionStatus = response['collectionStatus']

        return None

    def getInterfaces(self):
        endpoint = f"interface/network-device/{self.id}"
        data = restcall('GET', self.dnac, endpoint)
        return data

    def getTopology(self):
        ret = []
        endpoint = "topology/physical-topology/"
        data = restcall('GET', self.dnac, endpoint)
        connections = {}
        links = []
        for link in data['response']['links']:

            try:
                connections['sourcenode'] = link['source']
                connections['remotenode'] = link['target']
                connections['sourceif'] = link['startPortName']
                connections['remoteif'] = link['endPortName']
                links.append(dict(connections))
            except:
                pass
        ret = links
        return ret

    def getConnections(self):
        ret = []
        endpoint = "topology/physical-topology/"
        data = restcall('GET', self.dnac, endpoint)
        connections = {}
        links = []
        for link in data['response']['links']:
            try:
                if link['source'] == self.id:
                    connections['remotenode'] = link['target']
                    connections['remoteif'] = link['endPortName']
                    connections['localif'] = link['startPortName']
                    links.append(dict(connections))
                elif link['target'] == self.id:
                    connections['remotenode'] = link['source']
                    connections['remoteif'] = link['startPortName']
                    connections['localif'] = link['endPortName']
                    links.append(dict(connections))
            except:
                pass
        ret = links
        return ret

    def deployTemplate(self, template):
        """
        Inputs:
        template (obj) ezdnac template object

        Returns data (dict):
        {
        'deploymetId': id(str),   if error occurs, deploymentId returns None.
        'message': (str)
        }
        """
        data = {}
        data['deploymentId'] = None
        data['message'] = None
        endpoint = "template-programmer/template/deploy"

        payload = {
            "templateId": template.id,
            "targetInfo": [
                {
                    "id": self.id,
                    "type": "MANAGED_DEVICE_UUID",
                    "params": template.params
                }
            ]}

        response = restcall('POST', self.dnac, endpoint, jsondata=payload)

        # If error occurs, no id
        if 'response' in response and 'errorCode' in response['response']:
            data['deploymentId'] = None
            data['message'] = response['response']
            data['deployed'] = False
            return data

        # This is how the id should be found, if the API wasnt broken.
        if 'deploymentId' in response and type(response['deploymentId']) is int:
            data['deploymentId'] = response['deploymentId']
            data['message'] = 'Id found in response'
            data['deployed'] = True
            self.deploymentId = response
            return data

        # This is how it is solved with regex instead.
        if 'deploymentId' in response and type(response['deploymentId']) is not int:
            data['message'] = 'Id found with regex, broken response'

            # Testing if id is in response:
            resultRegex = (re.findall(r'Template Deployemnt Id.*', response['deploymentId']))

            if len(resultRegex) is not 0:
                Id = str(resultRegex).strip("['Template Deployemnt Id: ]")
                if resultRegex is not None:
                    data['deploymentId'] = Id
                    data['deployed'] = True
                    self.deploymentId = data
                    return data

        # If template was not deployed
        if re.match(r'.*already deployed with same params.*', response['deploymentId']):
            data['deploymentId'] = None
            data['deployed'] = True
            data['message'] = 'Same version already deployed with same params'
        else:
            data['deploymentId'] = None
            data['deployed'] = False
            data['message'] = response

        return data
 

    def deployTemplateStatus(self, **kwargs):
        if 'id' in kwargs:
            self.deploymentId = kwargs['id']

        if self.deploymentId == None:
            return None

        if self.deploymentId['deploymentId'] == None:
            return self.deploymentId['message']
        else:
            endpoint = f"template-programmer/template/deploy/status/{self.deploymentId['deploymentId']}"
            data = restcall('GET', self.dnac, endpoint)

            if 'status' in data:
                return data['status']
            else:
                return data

    def deployTemplateReport(self, **kwargs):
        if 'id' in kwargs:
            self.deploymentId = kwargs['id']        

        endpoint = f"template-programmer/template/deploy/status/{self.deploymentId['deploymentId']}"
        response = restcall('GET', self.dnac, endpoint)
        return response

    def findNextPortchannel(self):
        endpoint = f"interface/network-device/{self.id}"
        response = restcall('GET', self.dnac, endpoint)

        existing_ids = []
        for interface in response['response']:
            if re.match(r'Port-channel.*', str(interface['portName'])):
                intf = int(str(interface['portName']).strip("'Port-channel"))
                existing_ids.append(intf)

        for i in range(1, 49):
            if (i) not in existing_ids:
                next_id = i
                break
        return next_id

    def assignToSite(self, siteId):
        baseurl = "/dna/system/api/v1/"
        endpoint = f"site/{siteId}/device"
        payload = {
            "device": [
                {
                    "ip": self.ip
                }
            ]
        }

        response = restcall('POST', self.dnac, endpoint, baseurl=baseurl, jsondata=payload)
        try:
            self.executionId = response['executionId']
        except:
            pass
        return response

    def getNeighbors(self):
        connections = self.getConnections()
        neighbors = []
        for link in connections:
            if link['remotenode'] in neighbors:
                continue
            else:
                neighbors.append(link['remotenode'])
        return neighbors

    # return every interface connected to us from specific neighbor

    def getNeighborIfs(self, neighbor):
        connections = self.getConnections()
        interfaces = []
        for link in connections:
            if link['remotenode'] == neighbor:
                interfaces.append(link['remoteif'])
        return interfaces

    def getModules(self):
        endpoint = f'network-device/module?deviceId={self.id}'
        data = restcall('GET', self.dnac, endpoint)

        modules = data['modules']
        self.modules = modules

        switches = []
        for module in modules:
            name = module['name']
            switch = str((re.findall(r'Switch \d', name))).strip("[']")
            switches.append(switch)
        self.stackcount = len((set(switches)))

        return modules

    def claimDevice(self, siteId, **kwargs):
        endpoint = "onboarding/pnp-device/site-claim"
        
        if 'template' in kwargs:
            payload = {
                "siteId": siteId,
                "deviceId": self.id,
                "type": "Default",
                "imageInfo": {"imageId": "None", "skip": "true"},
                "configInfo": {"configId": "", "configParameters": [kwargs['template'].params]}
            }
        else:
            payload = {
                "siteId": siteId,
                "deviceId": self.id,
                "type": "Default",
                "imageInfo": {"imageId": "None", "skip": "true"},
                "configInfo": {"configId": "", "configParameters": []}
            }

        print(payload)
        data = restcall('POST', self.dnac, endpoint, jsondata=payload)
        print(data)
        try:
            self.executionId = response['executionId']
        except:
            pass
        return data

    def sync(self):
        baseurl = '/dna/intent/api/v1/'
        endpoint = 'network-device/sync'
        payload = f'["{self.id}"]\n\n'
        data = restcall('PUT', self.dnac, endpoint, data=payload, baseurl=baseurl)

        self.taskId = data['response']['taskId']
        return data

    def getTaskStatus(self, **kwargs):
        if 'id' in kwargs:
            taskId = kwargs['id']
        elif self.taskId is not None:
            taskId = self.taskId
        else:
            return None

        endpoint = f'task/{self.taskId}'
        data = restcall('GET', self.dnac, endpoint)
        return data

    def getExecutionStatus(self, **kwargs):
        if 'id' in kwargs:
            executionId = kwargs['id']
        elif self.executionId is not None:
            executionId = self.executionId
        else:
            return None

        baseurl = '/dna/platform/management/business-api/v1/'
        endpoint = f'execution-status/{self.executionId}'
        data = restcall('GET', self.dnac, endpoint, baseurl=baseurl)
        return data
